@@ -63,6 +63,80 @@ struct Gsasl_session
 #endif
 };
 
+/* another internal GSASL structures needed to hack into its internals */
+#define DIGEST_MD5_LENGTH 16
+
+struct digest_md5_challenge
+{
+  size_t nrealms;
+  char **realms;
+  char *nonce;
+  int qops;
+  int stale;
+  unsigned long servermaxbuf;
+  int utf8;
+  int ciphers;
+};
+typedef struct digest_md5_challenge digest_md5_challenge;
+
+enum digest_md5_qop
+{
+  DIGEST_MD5_QOP_AUTH = 1,
+  DIGEST_MD5_QOP_AUTH_INT = 2,
+  DIGEST_MD5_QOP_AUTH_CONF = 4
+};
+typedef enum digest_md5_qop digest_md5_qop;
+
+enum digest_md5_cipher
+{
+  DIGEST_MD5_CIPHER_DES = 1,
+  DIGEST_MD5_CIPHER_3DES = 2,
+  DIGEST_MD5_CIPHER_RC4 = 4,
+  DIGEST_MD5_CIPHER_RC4_40 = 8,
+  DIGEST_MD5_CIPHER_RC4_56 = 16,
+  DIGEST_MD5_CIPHER_AES_CBC = 32
+};
+typedef enum digest_md5_cipher digest_md5_cipher;
+
+#define DIGEST_MD5_RESPONSE_LENGTH 32
+struct digest_md5_response
+{
+  char *username;
+  char *realm;
+  char *nonce;
+  char *cnonce;
+  unsigned long nc;
+  digest_md5_qop qop;
+  char *digesturi;
+  unsigned long clientmaxbuf;
+  int utf8;
+  digest_md5_cipher cipher;
+  char *authzid;
+  char response[DIGEST_MD5_RESPONSE_LENGTH + 1];
+};
+typedef struct digest_md5_response digest_md5_response;
+
+struct digest_md5_finish
+{
+  char rspauth[DIGEST_MD5_RESPONSE_LENGTH + 1];
+};
+typedef struct digest_md5_finish digest_md5_finish;
+
+struct _Gsasl_digest_md5_server_state
+{
+  int step;
+  unsigned long readseqnum, sendseqnum;
+  char secret[DIGEST_MD5_LENGTH];
+  char kic[DIGEST_MD5_LENGTH];
+  char kcc[DIGEST_MD5_LENGTH];
+  char kis[DIGEST_MD5_LENGTH];
+  char kcs[DIGEST_MD5_LENGTH];
+  digest_md5_challenge challenge;
+  digest_md5_response response;
+  digest_md5_finish finish;
+};
+typedef struct _Gsasl_digest_md5_server_state _Gsasl_digest_md5_server_state;
+
 /** utility: generate a success nad */
 static nad_t _sx_sasl_success(sx_t s) {
     nad_t nad;
@@ -134,6 +208,60 @@ static nad_t _sx_sasl_abort(sx_t s) {
     return nad;
 }
 
+static int _sx_sasl_wio(sx_t s, sx_plugin_t p, sx_buf_t buf) {
+    sx_error_t sxe;
+    int len, ret;
+    char *out;
+    Gsasl_session *sd = (Gsasl_session *) s->plugin_data[p->index];
+
+    _sx_debug(ZONE, "doing sasl encode");
+
+    /* encode the output */
+    ret = gsasl_encode(sd, buf->data, buf->len, &out, &len);
+    if (ret != GSASL_OK) {
+        _sx_debug(ZONE, "gsasl_encode failed (%d): %s", ret, gsasl_strerror (ret));
+        /* Fatal error */
+        _sx_gen_error(sxe, SX_ERR_AUTH, "SASL Stream encoding failed", (char*) gsasl_strerror (ret));
+        _sx_event(s, event_ERROR, (void *) &sxe);
+        return -1;
+    }
+    
+    /* replace the buffer */
+    _sx_buffer_set(buf, out, len, NULL);
+    free(out);
+
+    _sx_debug(ZONE, "%d bytes encoded for sasl channel", buf->len);
+    
+    return 1;
+}
+
+static int _sx_sasl_rio(sx_t s, sx_plugin_t p, sx_buf_t buf) {
+    sx_error_t sxe;
+    int len, ret;
+    char *out;
+    Gsasl_session *sd = (Gsasl_session *) s->plugin_data[p->index];
+
+    _sx_debug(ZONE, "doing sasl decode");
+
+    /* decode the input */
+    ret = gsasl_decode(sd, buf->data, buf->len, &out, &len);
+    if (ret != GSASL_OK) {
+        _sx_debug(ZONE, "gsasl_decode failed (%d): %s", ret, gsasl_strerror (ret));
+        /* Fatal error */
+        _sx_gen_error(sxe, SX_ERR_AUTH, "SASL Stream decoding failed", (char*) gsasl_strerror (ret));
+        _sx_event(s, event_ERROR, (void *) &sxe);
+        return -1;
+    }
+    
+    /* replace the buffer */
+    _sx_buffer_set(buf, out, len, NULL);
+    free(out);
+
+    _sx_debug(ZONE, "%d bytes decoded from sasl channel", len);
+    
+    return 1;
+}
+
 /** move the stream to the auth state */
 void _sx_sasl_open(sx_t s, Gsasl_session *sd) {
     char *method, *authzid;
@@ -260,6 +388,9 @@ static void _sx_sasl_features(sx_t s, sx_plugin_t p, nad_t nad) {
 
 /** auth done, restart the stream */
 static void _sx_sasl_notify_success(sx_t s, void *arg) {
+    sx_plugin_t p = (sx_plugin_t) arg;
+
+    _sx_chain_io_plugin(s, p);
     _sx_debug(ZONE, "auth completed, resetting");
 
     _sx_reset(s);
@@ -300,6 +431,12 @@ static void _sx_sasl_client_process(sx_t s, sx_plugin_t p, Gsasl_session *sd, ch
         gsasl_property_set(sd, GSASL_SERVICE, ctx->appname);
         gsasl_property_set(sd, GSASL_REALM, realm);
 
+        /* allow only qop=auth for DIGEST-MD5 */
+        if (!strncmp(mech, "DIGEST-MD5", 10)) {
+            _Gsasl_digest_md5_server_state *state = sd->mech_data;
+            state->challenge.qops = DIGEST_MD5_QOP_AUTH;
+        }
+
         /* get hostname */
         hostname[0] = '\0';
         gethostname(hostname, 256);
@@ -339,7 +476,8 @@ static void _sx_sasl_client_process(sx_t s, sx_plugin_t p, Gsasl_session *sd, ch
             if (ret != GSASL_OK) {
                 _sx_debug(ZONE, "gsasl_base64_from failed, no sasl for this conn; (%d): %s", ret, gsasl_strerror(ret));
                 _sx_nad_write(s, _sx_sasl_failure(s, _sasl_err_INCORRECT_ENCODING), 0);
-				return;
+                if(buf != NULL) free(buf);
+                return;
             }
         }
 
@@ -396,7 +534,6 @@ static void _sx_sasl_client_process(sx_t s, sx_plugin_t p, Gsasl_session *sd, ch
 
         /* encode the challenge */
         ret = gsasl_base64_to(out, outlen, &buf, &buflen);
-        
         if (ret == GSASL_OK) {
             _sx_nad_write(s, _sx_sasl_challenge(s, buf, buflen), 0);
             free(buf);
@@ -418,7 +555,7 @@ static void _sx_sasl_client_process(sx_t s, sx_plugin_t p, Gsasl_session *sd, ch
 
 /** process handshake packets from the server */
 static void _sx_sasl_server_process(sx_t s, sx_plugin_t p, Gsasl_session *sd, char *in, int inlen) {
-    char *buf, *out;
+    char *buf = NULL, *out = NULL;
     size_t buflen, outlen;
     int ret;
 
@@ -432,7 +569,7 @@ static void _sx_sasl_server_process(sx_t s, sx_plugin_t p, Gsasl_session *sd, ch
     
         /* process the data */
         ret = gsasl_step(sd, buf, buflen, &out, &outlen);
-        if(buf != NULL) free(buf);
+        if(buf != NULL) free(buf); buf = NULL;
     
         /* in progress */
         if(ret == GSASL_OK || ret == GSASL_NEEDS_MORE) {
@@ -440,19 +577,19 @@ static void _sx_sasl_server_process(sx_t s, sx_plugin_t p, Gsasl_session *sd, ch
     
             /* encode the response */
             ret = gsasl_base64_to(out, outlen, &buf, &buflen);
-    
+
             if (ret == GSASL_OK) {
                 _sx_nad_write(s, _sx_sasl_response(s, buf, buflen), 0);
-                if(buf != NULL) free(buf);
             }
-    
+
             if(out != NULL) free(out);
+            if(buf != NULL) free(buf);
     
             return;
         }
-    
-        if(out != NULL) free(out);
     }
+    if(out != NULL) free(out);
+    if(buf != NULL) free(buf);
 
     /* its over */
     _sx_debug(ZONE, "sasl handshake aborted; (%d): %s", ret, gsasl_strerror(ret));
@@ -750,6 +887,8 @@ int sx_sasl_init(sx_env_t env, sx_plugin_t p, va_list args) {
     p->private = (void *) ctx;
 
     p->unload = _sx_sasl_unload;
+    p->wio = _sx_sasl_wio;
+    p->rio = _sx_sasl_rio;
 
     p->stream = _sx_sasl_stream;
     p->features = _sx_sasl_features;
@@ -764,7 +903,7 @@ int sx_sasl_init(sx_env_t env, sx_plugin_t p, va_list args) {
 int sx_sasl_auth(sx_plugin_t p, sx_t s, char *appname, char *mech, char *user, char *pass) {
     _sx_sasl_t ctx = (_sx_sasl_t) p->private;
     Gsasl_session *sd;
-    char *buf, *out;
+    char *buf = NULL, *out = NULL;
     char hostname[256];
     int ret, ns;
     size_t buflen, outlen;
@@ -825,6 +964,7 @@ int sx_sasl_auth(sx_plugin_t p, sx_t s, char *appname, char *mech, char *user, c
 
         gsasl_finish(sd);
 
+        if (out != NULL) free(out);
         return 1;
     }
     free(out);
